//! microrm is a simple object relational manager (ORM) for sqlite.
//!
//! Unlike many fancier ORM systems, microrm is designed to be lightweight, both in terms of
//! runtime overhead and developer LoC. By necessity, it sacrifices flexibility towards these
//! goals, and so can be thought of as more opinionated than, say,
//! [SeaORM](https://www.sea-ql.org/SeaORM/) or [Diesel](https://diesel.rs/). Major limitations of
//! microrm are:
//! - lack of database migration support
//! - limited vocabulary for describing object-to-object relations
//!
//! There are three externally-facing components in microrm:
//! - Object modelling (via the [`Datum`](schema/datum/trait.Datum.html) and
//! [`Entity`](schema/entity/trait.Entity.html) traits)
//! - Database querying (via [`Queryable`](prelude/trait.Queryable.html),
//! [`AssocInterface`](prelude/trait.AssocInterface.html) and
//! [`Insertable`](prelude/trait.Insertable.html) traits)
//! - Command-line interface generation via the [`clap`](https://docs.rs/clap/latest/clap/) crate
//! (see [`cli::CLIObject`] and [`cli::ClapInterface`]; requires the optional crate feature `clap`)
//!
//! microrm pushes the Rust type system somewhat to provide better ergonomics, so the MSRV is
//! currently 1.75. Don't be scared off by the web of traits in the `schema` module --- you should
//! never need to interact with any of them!

// to make the proc_macros work inside the microrm crate; needed for tests and the metaschema.
extern crate self as microrm;

pub mod db;
mod query;
pub mod schema;

#[cfg(feature = "clap")]
pub mod cli;

pub mod prelude {
    pub use crate::query::{AssocInterface, Insertable, Queryable};
    // pub use crate::schema::entity::Entity;
    pub use crate::schema::{AssocMap, Database, IDMap};
    pub use microrm_macros::{Database, Entity};
}

// ----------------------------------------------------------------------
// Generically-useful database types
// ----------------------------------------------------------------------

#[derive(Debug)]
pub enum Error {
    EmptyResult,
    UnknownValue(String),
    IncompatibleSchema,
    InternalError(&'static str),
    EncodingError(std::str::Utf8Error),
    LogicError(&'static str),
    ConstraintViolation(&'static str),
    Sqlite {
        code: i32,
        msg: String,
        sql: Option<String>,
    },
    JSON(serde_json::Error),
    LockError(String),
}

impl std::fmt::Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        <Self as std::fmt::Debug>::fmt(self, f)
    }
}

impl<T> From<std::sync::PoisonError<T>> for Error {
    fn from(value: std::sync::PoisonError<T>) -> Self {
        Self::LockError(value.to_string())
    }
}

impl From<std::ffi::NulError> for Error {
    fn from(_value: std::ffi::NulError) -> Self {
        Self::InternalError("NULL pointer encountered in unexpected location")
    }
}

impl From<std::str::Utf8Error> for Error {
    fn from(value: std::str::Utf8Error) -> Self {
        Self::EncodingError(value)
    }
}

impl std::error::Error for Error {}

pub type DBResult<T> = Result<T, Error>;
