//! Component types for query construction.

use crate::{
    db::{Connection, StatementContext},
    prelude::Queryable,
    query::{AssocInterface, QueryPart},
    schema::{
        datum::{Datum, DatumDiscriminator, DatumVisitor, QueryEquivalent, QueryEquivalentList},
        entity::{Entity, EntityPart, EntityPartList, EntityPartVisitor},
        LocalSide, Stored,
    },
};

use super::Query;

/// Allow manipulation of an entire table.
pub(crate) struct TableComponent<E: Entity> {
    conn: Connection,
    _ghost: std::marker::PhantomData<E>,
}

impl<E: Entity> Clone for TableComponent<E> {
    fn clone(&self) -> Self {
        Self {
            conn: self.conn.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<E: Entity> TableComponent<E> {
    pub fn new(conn: Connection) -> Self {
        Self {
            conn,
            _ghost: Default::default(),
        }
    }
}

impl<E: Entity> Queryable for TableComponent<E> {
    type EntityOutput = E;
    type OutputContainer = Vec<Stored<E>>;
    type StaticVersion = Self;

    fn build(&self) -> Query {
        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT".into())
            .attach(QueryPart::Columns, "*".into())
            .attach(QueryPart::From, format!("`{}`", E::entity_name()))
    }
    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {}
    fn conn(&self) -> &Connection {
        &self.conn
    }
}

/// Filter on a Datum
#[derive(Clone)]
pub(crate) struct WithComponent<WEP: EntityPart, Parent: Queryable, QE: QueryEquivalent<WEP::Datum>>
{
    datum: QE,
    parent: Parent,
    _ghost: std::marker::PhantomData<WEP>,
}

impl<WEP: EntityPart, Parent: Queryable, QE: QueryEquivalent<WEP::Datum>>
    WithComponent<WEP, Parent, QE>
{
    pub fn new(parent: Parent, _part: WEP, datum: QE) -> Self {
        Self {
            datum,
            parent,
            _ghost: Default::default(),
        }
    }
}

/// this workaround is needed because we very explicitly would like EL to not be a
/// 'static-restricted type, and it doesn't matter for the purposes of actually distinguishing
/// between queries.
#[derive(Clone)]
pub(crate) struct CanonicalWithComponent<WEP: EntityPart, Parent: Queryable> {
    _ghost: std::marker::PhantomData<(WEP, Parent)>,
}

impl<WEP: EntityPart, Parent: Queryable + 'static> Queryable
    for CanonicalWithComponent<WEP, Parent>
{
    type EntityOutput = WEP::Entity;
    type OutputContainer = Option<Stored<WEP::Entity>>;
    type StaticVersion = Self;

    fn build(&self) -> Query {
        unreachable!()
    }
    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {
        unreachable!()
    }
    fn conn(&self) -> &Connection {
        unreachable!()
    }
}

impl<
        Parent: Queryable,
        WEP: EntityPart<Entity = Parent::EntityOutput>,
        QE: QueryEquivalent<WEP::Datum>,
    > Queryable for WithComponent<WEP, Parent, QE>
{
    type EntityOutput = WEP::Entity;
    type OutputContainer = Parent::OutputContainer;
    type StaticVersion = CanonicalWithComponent<WEP, Parent::StaticVersion>;

    fn build(&self) -> Query {
        self.parent.build().attach(
            QueryPart::Where,
            format!(
                "`{}`.`{}` = ?",
                <Self::EntityOutput>::entity_name(),
                WEP::part_name()
            ),
        )
    }
    fn bind(&self, stmt: &mut StatementContext, index: &mut i32) {
        self.parent.bind(stmt, index);
        self.datum.bind_to(stmt, *index);
        *index += 1;
    }

    fn conn(&self) -> &Connection {
        self.parent.conn()
    }
}

/// Filter on the keying index
pub(crate) struct KeyComponent<
    E: Entity,
    Parent: Queryable,
    EL: QueryEquivalentList<<E::Keys as EntityPartList>::DatumList>,
> {
    datum: EL,
    parent: Parent,
    _ghost: std::marker::PhantomData<E>,
}

impl<
        E: Entity,
        Parent: Queryable,
        EL: QueryEquivalentList<<E::Keys as EntityPartList>::DatumList>,
    > KeyComponent<E, Parent, EL>
{
    pub fn new(parent: Parent, datum: EL) -> Self {
        Self {
            datum,
            parent,
            _ghost: Default::default(),
        }
    }
}

impl<
        E: Entity,
        Parent: Queryable,
        EL: QueryEquivalentList<<E::Keys as EntityPartList>::DatumList>,
    > Clone for KeyComponent<E, Parent, EL>
{
    fn clone(&self) -> Self {
        Self {
            datum: self.datum.clone(),
            parent: self.parent.clone(),
            _ghost: Default::default(),
        }
    }
}

/// this workaround is needed because we very explicitly would like EL to not be a
/// 'static-restricted type, and it doesn't matter for the purposes of actually distinguishing
/// between queries.
pub(crate) struct CanonicalUniqueComponent<E: Entity, Parent: Queryable> {
    _ghost: std::marker::PhantomData<(E, Parent)>,
}

impl<E: Entity, Parent: Queryable + 'static> Queryable for CanonicalUniqueComponent<E, Parent> {
    type EntityOutput = E;
    type OutputContainer = Option<Stored<E>>;
    type StaticVersion = Self;

    fn build(&self) -> Query {
        unreachable!()
    }
    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {
        unreachable!()
    }
    fn conn(&self) -> &Connection {
        unreachable!()
    }
}

impl<E: Entity, Parent: Queryable + 'static> Clone for CanonicalUniqueComponent<E, Parent> {
    fn clone(&self) -> Self {
        Self {
            _ghost: Default::default(),
        }
    }
}

impl<
        E: Entity,
        Parent: Queryable,
        EL: QueryEquivalentList<<E::Keys as EntityPartList>::DatumList>,
    > Queryable for KeyComponent<E, Parent, EL>
{
    type EntityOutput = E;
    type OutputContainer = Option<Stored<E>>;
    type StaticVersion = CanonicalUniqueComponent<E, Parent::StaticVersion>;

    fn build(&self) -> Query {
        let mut query = self.parent.build();

        struct PartVisitor<'a>(&'a mut Query);
        impl<'a> EntityPartVisitor for PartVisitor<'a> {
            fn visit<EP: EntityPart>(&mut self) {
                self.0.attach_mut(
                    QueryPart::Where,
                    format!(
                        "`{}`.`{}` = ?",
                        <EP::Entity>::entity_name(),
                        EP::part_name()
                    ),
                );
            }
        }

        <E::Keys>::accept_part_visitor(&mut PartVisitor(&mut query));

        query
    }

    fn bind(&self, stmt: &mut StatementContext, index: &mut i32) {
        self.parent.bind(stmt, index);

        struct Visitor<'a, 'b>(&'a mut StatementContext<'b>, &'a mut i32);
        impl<'a, 'b> DatumVisitor for Visitor<'a, 'b> {
            fn visit<ED: Datum>(&mut self, datum: &ED) {
                datum.bind_to(self.0, *self.1);
                *self.1 += 1;
            }
        }

        self.datum.accept(&mut Visitor(stmt, index));
    }

    fn conn(&self) -> &Connection {
        self.parent.conn()
    }
}

#[derive(Clone)]
pub(crate) struct SingleComponent<Parent: Queryable> {
    parent: Parent,
}

impl<Parent: Queryable> SingleComponent<Parent> {
    pub fn new(parent: Parent) -> Self {
        Self { parent }
    }
}

impl<Parent: Queryable> Queryable for SingleComponent<Parent> {
    type EntityOutput = Parent::EntityOutput;
    type OutputContainer = Option<Stored<Self::EntityOutput>>;
    type StaticVersion = SingleComponent<Parent::StaticVersion>;

    fn build(&self) -> Query {
        self.parent
            .build()
            .attach(QueryPart::Trailing, "LIMIT 1".into())
    }

    fn bind(&self, stmt: &mut StatementContext, index: &mut i32) {
        self.parent.bind(stmt, index)
    }

    fn conn(&self) -> &Connection {
        self.parent.conn()
    }
}

/// Join with another entity via an association
pub(crate) struct JoinComponent<R: Entity, L: Entity, EP: EntityPart<Entity = L>, Parent: Queryable>
{
    parent: Parent,
    _ghost: std::marker::PhantomData<(L, R, EP)>,
}

impl<R: Entity, L: Entity, EP: EntityPart<Entity = L>, Parent: Queryable>
    JoinComponent<R, L, EP, Parent>
{
    pub fn new(parent: Parent, _part: EP) -> Self {
        Self {
            parent,
            _ghost: Default::default(),
        }
    }
}

impl<R: Entity, L: Entity, EP: EntityPart<Entity = L>, Parent: Queryable> Clone
    for JoinComponent<R, L, EP, Parent>
{
    fn clone(&self) -> Self {
        Self {
            parent: self.parent.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<
        R: Entity,
        L: Entity,
        EP: EntityPart<Entity = L, Datum = AI>,
        AI: AssocInterface + Datum,
        Parent: Queryable,
    > Queryable for JoinComponent<R, L, EP, Parent>
{
    type EntityOutput = R;
    type OutputContainer = Vec<Stored<R>>;
    type StaticVersion = JoinComponent<R, L, EP, Parent::StaticVersion>;

    fn build(&self) -> Query {
        let remote_name = R::entity_name();
        let local_name = L::entity_name();

        struct Discriminator(Option<(&'static str, &'static str)>, Option<&'static str>);
        impl DatumDiscriminator for Discriminator {
            fn visit_entity_id<E: Entity>(&mut self) {
                unreachable!()
            }
            fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(&mut self) {
                unreachable!()
            }
            fn visit_bare_field<T: Datum>(&mut self) {
                unreachable!()
            }

            fn visit_assoc_map<E: Entity>(&mut self) {
                self.0 = Some(("domain", "range"));
            }
            fn visit_assoc_domain<R: crate::schema::Relation>(&mut self) {
                self.0 = Some(("domain", "range"));
                self.1 = Some(R::NAME);
            }
            fn visit_assoc_range<R: crate::schema::Relation>(&mut self) {
                self.0 = Some(("range", "domain"));
                self.1 = Some(R::NAME);
            }
        }

        let mut d = Discriminator(None, None);
        <EP::Datum>::accept_discriminator(&mut d);

        let (local_field, remote_field) = d.0.unwrap();

        // table name format is {domain}_{range}_assoc_{name}
        let assoc_name = match AI::SIDE {
            LocalSide::Range => {
                format!(
                    "{remote_name}_{local_name}_assoc_{}",
                    d.1.unwrap_or(EP::part_name())
                )
            }
            LocalSide::Domain => {
                format!(
                    "{local_name}_{remote_name}_assoc_{}",
                    d.1.unwrap_or(EP::part_name())
                )
            }
        };

        self.parent
            .build()
            .attach(
                QueryPart::Join,
                format!("`{assoc_name}` ON `{local_name}`.`id` = `{assoc_name}`.`{local_field}`"),
            )
            .attach(
                QueryPart::Join,
                format!(
                    "`{remote_name}` ON `{assoc_name}`.`{remote_field}` = `{remote_name}`.`id`"
                ),
            )
            .replace(QueryPart::Columns, format!("`{remote_name}`.*"))
    }
    fn bind(&self, stmt: &mut StatementContext, index: &mut i32) {
        self.parent.bind(stmt, index);
    }

    fn conn(&self) -> &Connection {
        self.parent.conn()
    }
}

/*
SQL mapping:

MapQueryable<E>::get()
 -> SELECT * FROM {E::entity_name()}

UniqueComponent<E, MapQueryable<E>>::get()
 -> SELECT * FROM {E::entity_name()} WHERE `col1` = ? AND `col2` = ? AND ...

WithComponent<E, EP, MapQueryable<E>>::get()
 -> SELECT * FROM {E::entity_name()} WHERE `col1` = ?

AssocQueryable<AI>::get()
 -> SELECT
        `{AI::RemoteEntity::entity_name()}`.*
    FROM `{assoc_table_name}`
    LEFT JOIN `{AI::RemoteEntity::entity_name()}`.id = `{assoc_table_name}`.`{remote_field}`
    WHERE `{assoc_table_name}`.`{local_field}` = ?

UniqueComponent<E, AssocQueryable<AI>>::get()
 -> SELECT
        `{AI::RemoteEntity::entity_name()}`.*
    FROM `{assoc_table_name}`
    LEFT JOIN `{AI::RemoteEntity::entity_name()}`.id = `{assoc_table_name}`.`{remote_field}`
    WHERE
        `{assoc_table_name}`.`{local_field}` = ?
        AND `{AI::RemoteEntity::entity_name()}`.`col1` = ?
        AND `{AI::RemoteEntity::entity_name()}`.`col2` = ?
        ...

JoinComponent<R, E, EP, MapQueryable<E>>::get()
 -> SELECT DISTINCT
        `{R::entity_name()}`.*
    FROM
        `{E::entity_name()}`
    LEFT JOIN `{assoc_table_name}` ON `{E::entity_name()}`.`id` = `{assoc_table_name}`.`{local_field}`
    LEFT JOIN `{R::entity_name()}` ON `{assoc_table_name}`.`{remote_field}` = `{R::entity_name()}`.`id`

*/
