//! Schema specification
//!
//! Terminology used:
//! - domain: one side of an association/relation, the "pointed-from" side for one-sided relations
//! - range: one side of an association/relation, the "pointed-to" side for one-sided relations
//! - local: the current side of an association
//! - remote: the opposite side of an association

use query::Queryable;

use crate::{
    db::{Connection, StatementContext, StatementRow, Transaction},
    query::{self, AssocInterface, Insertable},
    schema::datum::{Datum, DatumDiscriminator},
    schema::entity::{Entity, EntityVisitor},
};
use crate::{DBResult, Error};

use self::datum::{ConcreteDatum, DatumDiscriminatorRef};

pub mod datum;
pub mod entity;

mod build;
mod collect;
pub(crate) mod meta;
#[cfg(test)]
mod tests;

// ----------------------------------------------------------------------
// API types
// ----------------------------------------------------------------------

/// Wrapper struct that holds both an EntityID and an Entity itself.
pub struct Stored<T: Entity> {
    db: Connection,
    id: T::ID,
    wrap: T,
}

impl<T: Entity> Stored<T> {
    pub(crate) fn new(db: Connection, id: T::ID, value: T) -> Self {
        Self {
            db,
            id,
            wrap: value,
        }
    }

    pub fn id(&self) -> T::ID {
        self.id
    }

    pub fn wrapped(self) -> T {
        self.wrap
    }

    pub fn sync(&mut self) -> DBResult<()> {
        let txn = Transaction::new(&self.db)?;
        query::update_entity(&self.db, self)?;
        txn.commit()
    }
}

impl<T: Entity + std::fmt::Debug> std::fmt::Debug for Stored<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "Stored {{ id: {:?}, value: {:?} }}",
            self.id, self.wrap
        ))
    }
}

impl<T: Entity> AsRef<T> for Stored<T> {
    fn as_ref(&self) -> &T {
        &self.wrap
    }
}

impl<T: Entity> AsMut<T> for Stored<T> {
    fn as_mut(&mut self) -> &mut T {
        &mut self.wrap
    }
}

impl<T: Entity> std::ops::Deref for Stored<T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.wrap
    }
}

impl<T: Entity> std::ops::DerefMut for Stored<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.wrap
    }
}

impl<T: Entity> PartialEq for Stored<T> {
    fn eq(&self, other: &Self) -> bool {
        self.id == other.id
    }
}

// ----------------------------------------------------------------------
// Entity field types
// ----------------------------------------------------------------------

/// Represents an arbitrary relation between two types of entities.
///
/// In its most generic form, stores an arbitrary set of `(Domain::ID, Range::ID)` pairs used to
/// define the relation. Can be restricted to be injective if this is desired. Doing so will incur
/// a small runtime cost, since an extra index must be maintained.
pub trait Relation: 'static {
    type Domain: Entity;
    type Range: Entity;

    /// If true, then each Range-type entity can only be referred to by a single Domain-type
    /// entity. This roughly translates to the relation being computationally-efficiently
    /// invertible for a two-way map.
    const INJECTIVE: bool = false;

    const NAME: &'static str;
}

#[derive(Debug)]
pub enum LocalSide {
    Domain,
    Range,
}

/// Opaque data structure used for constructing `Assoc{Map,Domain,Range}` instances.
#[derive(Clone)]
pub struct AssocData {
    pub(crate) conn: Connection,
    pub(crate) local_name: &'static str,
    pub(crate) part_name: &'static str,
    pub(crate) local_id: i64,
}

/// Represents a simple one-to-many non-injective entity relationship.
///
/// This is a 'shortcut' type that doesn't require defining a tag type that implements `Relation`.
pub struct AssocMap<T: Entity> {
    pub(crate) data: Option<AssocData>,
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> Clone for AssocMap<T> {
    fn clone(&self) -> Self {
        Self {
            data: self.data.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<T: Entity> std::fmt::Debug for AssocMap<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "AssocMap {{ id: {:?} }}",
            self.data.as_ref().map(|d| d.local_id)
        ))
    }
}

impl<T: Entity> Default for AssocMap<T> {
    fn default() -> Self {
        Self::empty()
    }
}

impl<T: Entity> AssocInterface for AssocMap<T> {
    type RemoteEntity = T;
    const SIDE: LocalSide = LocalSide::Domain;

    fn get_distinguishing_name(&self) -> DBResult<&'static str> {
        self.data
            .as_ref()
            .ok_or(Error::LogicError(
                "no distinguishing name for empty AssocMap",
            ))
            .map(|d| d.part_name)
    }

    fn get_data(&self) -> DBResult<&AssocData> {
        self.data
            .as_ref()
            .ok_or(Error::LogicError("Reading from unassigned AssocMap"))
    }
}

impl<T: Entity> AssocMap<T> {
    pub fn empty() -> Self {
        Self {
            data: None,
            _ghost: Default::default(),
        }
    }
}

impl<T: Entity> Datum for AssocMap<T> {
    fn sql_type() -> &'static str {
        unreachable!()
    }

    fn debug_field(&self, _field: &'static str, _fmt: &mut std::fmt::DebugStruct)
    where
        Self: Sized,
    {
    }

    fn accept_entity_visitor(v: &mut impl EntityVisitor) {
        v.visit::<T>();
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_assoc_map::<T>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_assoc_map::<T>(self);
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        unreachable!()
    }

    fn build_from(adata: AssocData, _stmt: &mut StatementRow, _index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(Self {
            data: Some(adata),
            _ghost: Default::default(),
        })
    }

    fn update_adata(&mut self, adata: AssocData) {
        self.data = Some(adata);
    }
}
impl<T: Entity> ConcreteDatum for AssocMap<T> {}

/// The domain part of a full Relation definition.
pub struct AssocDomain<R: Relation> {
    pub(crate) data: Option<AssocData>,
    _ghost: std::marker::PhantomData<R>,
}

impl<R: Relation> Clone for AssocDomain<R> {
    fn clone(&self) -> Self {
        Self {
            data: self.data.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<R: Relation> Default for AssocDomain<R> {
    fn default() -> Self {
        Self::empty()
    }
}

impl<R: Relation> std::fmt::Debug for AssocDomain<R> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "AssocDomain {{ id: {:?} }}",
            self.data.as_ref().map(|d| d.local_id)
        ))
    }
}

impl<R: Relation> AssocDomain<R> {
    pub fn empty() -> Self {
        Self {
            data: None,
            _ghost: Default::default(),
        }
    }
}

impl<R: Relation> AssocInterface for AssocDomain<R> {
    type RemoteEntity = R::Range;
    const SIDE: LocalSide = LocalSide::Domain;

    fn get_distinguishing_name(&self) -> DBResult<&'static str> {
        Ok(R::NAME)
    }

    fn get_data(&self) -> DBResult<&AssocData> {
        self.data
            .as_ref()
            .ok_or(Error::LogicError("Reading from unassigned AssocDomain"))
    }
}

impl<R: Relation> Datum for AssocDomain<R> {
    fn sql_type() -> &'static str {
        unreachable!()
    }

    fn debug_field(&self, _field: &'static str, _fmt: &mut std::fmt::DebugStruct)
    where
        Self: Sized,
    {
    }

    fn accept_entity_visitor(v: &mut impl EntityVisitor) {
        v.visit::<R::Range>();
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_assoc_domain::<R>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_assoc_domain(self);
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        unreachable!()
    }

    fn build_from(adata: AssocData, _stmt: &mut StatementRow, _index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(Self {
            data: Some(adata),
            _ghost: Default::default(),
        })
    }

    fn update_adata(&mut self, adata: AssocData) {
        self.data = Some(adata);
    }
}
impl<R: Relation> ConcreteDatum for AssocDomain<R> {}

/// The range part of a full Relation definition.
pub struct AssocRange<R: Relation> {
    pub(crate) data: Option<AssocData>,
    _ghost: std::marker::PhantomData<R>,
}

impl<R: Relation> Clone for AssocRange<R> {
    fn clone(&self) -> Self {
        Self {
            data: self.data.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<R: Relation> Default for AssocRange<R> {
    fn default() -> Self {
        Self::empty()
    }
}

impl<R: Relation> std::fmt::Debug for AssocRange<R> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "AssocRange {{ id: {:?} }}",
            self.data.as_ref().map(|d| d.local_id)
        ))
    }
}

impl<R: Relation> AssocRange<R> {
    pub fn empty() -> Self {
        Self {
            data: None,
            _ghost: Default::default(),
        }
    }
}

impl<R: Relation> AssocInterface for AssocRange<R> {
    type RemoteEntity = R::Domain;
    const SIDE: LocalSide = LocalSide::Range;

    fn get_distinguishing_name(&self) -> DBResult<&'static str> {
        Ok(R::NAME)
    }

    fn get_data(&self) -> DBResult<&AssocData> {
        self.data
            .as_ref()
            .ok_or(Error::LogicError("Reading from unassigned AssocRange"))
    }
}

impl<R: Relation> Datum for AssocRange<R> {
    fn sql_type() -> &'static str {
        unreachable!()
    }

    fn debug_field(&self, _field: &'static str, _fmt: &mut std::fmt::DebugStruct)
    where
        Self: Sized,
    {
    }

    fn accept_entity_visitor(v: &mut impl EntityVisitor) {
        v.visit::<R::Domain>();
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_assoc_range::<R>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_assoc_range(self);
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        unreachable!()
    }

    fn build_from(adata: AssocData, _stmt: &mut StatementRow, _index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(Self {
            data: Some(adata),
            _ghost: Default::default(),
        })
    }

    fn update_adata(&mut self, adata: AssocData) {
        self.data = Some(adata);
    }
}
impl<R: Relation> ConcreteDatum for AssocRange<R> {}

/// Stores an arbitrary Rust data type as serialized JSON in a string field.
#[derive(Clone)]
pub struct Serialized<T: serde::Serialize + serde::de::DeserializeOwned + Clone> {
    wrapped: T,
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Default + Clone> Default
    for Serialized<T>
{
    fn default() -> Self {
        Self {
            wrapped: T::default(),
        }
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone> std::fmt::Debug
    for Serialized<T>
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        <T as std::fmt::Debug>::fmt(&self.wrapped, f)
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> From<T> for Serialized<T> {
    fn from(value: T) -> Self {
        Self { wrapped: value }
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> AsRef<T> for Serialized<T> {
    fn as_ref(&self) -> &T {
        &self.wrapped
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> AsMut<T> for Serialized<T> {
    fn as_mut(&mut self) -> &mut T {
        &mut self.wrapped
    }
}

impl<T: 'static + serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone> Datum
    for Serialized<T>
{
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        let json = Box::pin(
            serde_json::to_string(&self.wrapped).expect("couldn't serialize object into JSON"),
        );

        <String as Datum>::bind_to(&*json.as_ref(), stmt, index);

        // transfer ownership so that the data is still around for later
        stmt.transfer(json);
    }

    fn build_from(adata: AssocData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let s = <String as Datum>::build_from(adata, stmt, index)?;

        let d = serde_json::from_str::<T>(s.as_str()).map_err(Error::JSON)?;

        Ok(Self { wrapped: d })
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_serialized::<T>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_serialized::<T>(&self.wrapped);
    }
}

impl<T: 'static + serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone>
    ConcreteDatum for Serialized<T>
{
}

// ----------------------------------------------------------------------
// Database specification types
// ----------------------------------------------------------------------

/// Table with EntityID-based lookup.
#[derive(Clone)]
pub struct IDMap<T: Entity> {
    pub(crate) conn: Connection,
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> IDMap<T> {
    pub fn build(db: Connection) -> Self {
        Self {
            conn: db,
            _ghost: std::marker::PhantomData,
        }
    }

    pub(crate) fn conn(&self) -> &Connection {
        &self.conn
    }

    /// Look up an Entity in this map by ID.
    pub fn by_id(&self, id: T::ID) -> DBResult<Option<Stored<T>>> {
        self.with(id, &id).first().get()
    }
}

impl<'l, T: Entity> Insertable<T> for IDMap<T> {
    /// Insert a new Entity into this map, and return its new ID.
    fn insert(&self, value: T) -> DBResult<T::ID> {
        let txn = Transaction::new(self.conn())?;
        let out = query::insert(self.conn(), &value)?;
        txn.commit()?;
        Ok(out)
    }

    fn insert_and_return(&self, value: T) -> DBResult<Stored<T>> {
        let txn = Transaction::new(self.conn())?;
        let out = query::insert_and_return(self.conn(), value)?;
        txn.commit()?;
        Ok(out)
    }
}

pub struct Index<T: Entity, Key: Datum> {
    _conn: Connection,
    _ghost: std::marker::PhantomData<(T, Key)>,
}

/// Represents a single root-level table or index in a database.
pub trait DatabaseItem {
    fn item_key() -> &'static str;
    fn dependency_keys() -> &'static [&'static str];

    fn is_index() -> bool {
        false
    }
    fn index_over() -> &'static str {
        unreachable!()
    }
    fn index_columns() -> &'static [&'static str] {
        unreachable!()
    }

    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

pub trait DatabaseItemVisitor {
    fn visit<DI: DatabaseItem>(&mut self)
    where
        Self: Sized;
}

pub trait DatabaseSpec {
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

impl<T: Entity> DatabaseSpec for IDMap<T> {
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor) {
        visitor.visit::<T>()
    }
}

/// A root structure for the database specification graph.
pub trait Database {
    fn open_path<U: AsRef<str>>(uri: U) -> DBResult<Self>
    where
        Self: Sized,
    {
        let conn = Connection::new(uri.as_ref())?;
        let schema = build::collect_from_database::<Self>();
        match schema.check(conn.clone()) {
            // schema checks out
            Some(true) => {}
            // schema doesn't match
            Some(false) => Err(Error::IncompatibleSchema)?,
            // no schema found
            None => {
                schema.create(conn.clone())?;
            }
        }

        conn.execute_raw_sql("PRAGMA foreign_keys = ON")?;

        Ok(Self::build(conn))
    }

    #[doc(hidden)]
    fn build(conn: Connection) -> Self
    where
        Self: Sized;

    fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor)
    where
        Self: Sized;
}
