use topological_sort::TopologicalSort;

use crate::{
    prelude::*,
    schema::{
        collect::{EntityStateContainer, PartType},
        meta, Connection, Database, DatabaseItem, DatabaseItemVisitor,
    },
    DBResult,
};

#[derive(Debug)]
struct ColumnInfo {
    name: &'static str,
    ty: String,
    fkey: Option<String>,
    unique: bool,
}

#[derive(Debug)]
struct TableInfo {
    table_name: String,
    columns: Vec<ColumnInfo>,
    constraints: Vec<String>,
    dependencies: Vec<String>,
}

impl TableInfo {
    fn new(name: String) -> Self {
        TableInfo {
            table_name: name,
            columns: vec![],
            constraints: vec![],
            dependencies: vec![],
        }
    }

    fn build_creation_query(&self) -> String {
        let columns = self.columns.iter().map(|col| {
            format!(
                ", `{}` {}{}",
                col.name,
                col.ty,
                if col.unique { " unique" } else { "" }
            )
        });
        let fkeys = self.columns.iter().filter_map(|col| {
            Some(format!(
                ", foreign key(`{}`) references {} on delete cascade",
                col.name,
                col.fkey.as_ref()?
            ))
        });

        format!(
            "create table `{}` (`id` integer primary key{}{}{});",
            self.table_name,
            columns.collect::<String>(),
            fkeys.collect::<String>(),
            self.constraints
                .iter()
                .fold(String::new(), |a, b| format!("{}, {}", a, b))
        )
    }
}

pub(crate) struct DatabaseSchema {
    signature: u64,
    queries: Vec<String>,
}

impl DatabaseSchema {
    const SCHEMA_SIGNATURE_KEY: &'static str = "schema_signature";

    /// Three possible results:
    /// - yes, this is a schema match (true)
    /// - no, this is not a schema match (false)
    /// - there is no schema that we know of (None)
    pub fn check(&self, db: Connection) -> Option<bool> {
        // attempt to use connection as a MetadataDB database
        let metadb = meta::MetadataDB::build(db);

        // check to see if the signature exists and matches
        metadb
            .metastore
            .keyed(&Self::SCHEMA_SIGNATURE_KEY.to_string())
            .get()
            .ok()
            .flatten()
            .map(|kv| kv.value.parse::<u64>().unwrap_or(0) == self.signature)
    }

    pub fn create(&self, db: Connection) -> DBResult<()> {
        db.execute_raw_sql("BEGIN TRANSACTION")?;
        for query in self.queries.iter() {
            log::trace!("Running creation query {query}");
            db.execute_raw_sql(query)?;
        }

        // attempt to use connection as a MetadataDB database
        let metadb = meta::MetadataDB::build(db.clone());

        for query in collect_from_database::<meta::MetadataDB>().queries.iter() {
            db.execute_raw_sql(query)?;
        }

        db.execute_raw_sql("COMMIT")?;

        // store signature
        metadb.metastore.insert(meta::Meta {
            key: Self::SCHEMA_SIGNATURE_KEY.into(),
            value: format!("{}", self.signature),
        })?;

        Ok(())
    }
}

pub(crate) fn collect_from_database<DB: Database>() -> DatabaseSchema {
    struct IV(EntityStateContainer);

    impl DatabaseItemVisitor for IV {
        fn visit<DI: DatabaseItem>(&mut self)
        where
            Self: Sized,
        {
            DI::accept_entity_visitor(&mut self.0.make_context());
        }
    }

    let mut iv = IV(EntityStateContainer::default());

    DB::accept_item_visitor(&mut iv);

    // now to turn all that into a set of tables
    let mut tables = std::collections::HashMap::new();

    for state in iv.0.iter_states() {
        let table_name = state.name.to_string();
        // we may end up visiting duplicate entities; skip them if so
        if tables.contains_key(&table_name) {
            continue;
        }

        let mut table = TableInfo::new(table_name.clone());
        for part in state.parts.iter() {
            match &part.ty {
                PartType::Datum(dtype) => table.columns.push(ColumnInfo {
                    name: part.name,
                    ty: dtype.to_string(),
                    fkey: None,
                    unique: part.unique,
                }),
                PartType::IDReference(entity_name) => table.columns.push(ColumnInfo {
                    name: part.name,
                    ty: "int".into(),
                    fkey: Some(format!("`{}`(`id`)", entity_name)),
                    unique: part.unique,
                }),
                PartType::AssocDomain {
                    table_name: assoc_table_name,
                    range_name,
                } => {
                    let mut assoc_table = TableInfo::new(assoc_table_name.clone());
                    assoc_table.dependencies.push(table_name.clone());
                    assoc_table.dependencies.push(range_name.to_string());

                    assoc_table.columns.push(ColumnInfo {
                        name: "domain",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", table_name)),
                        unique: false,
                    });

                    assoc_table.columns.push(ColumnInfo {
                        name: "range",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", range_name)),
                        unique: false,
                    });

                    assoc_table
                        .constraints
                        .push(format!("unique(`range`, `domain`)"));

                    tables.insert(assoc_table_name.clone(), assoc_table);
                }
                PartType::AssocRange {
                    table_name: assoc_table_name,
                    domain_name,
                } => {
                    let mut assoc_table = TableInfo::new(assoc_table_name.clone());
                    assoc_table.dependencies.push(table_name.clone());
                    assoc_table.dependencies.push(domain_name.to_string());

                    assoc_table.columns.push(ColumnInfo {
                        name: "domain",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", domain_name)),
                        unique: false,
                    });

                    assoc_table.columns.push(ColumnInfo {
                        name: "range",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", table_name)),
                        unique: false,
                    });

                    assoc_table
                        .constraints
                        .push(format!("unique(`range`, `domain`)"));
                    tables.insert(assoc_table_name.clone(), assoc_table);
                }
            }
        }

        let key = state.parts.iter().filter(|p| p.key).collect::<Vec<_>>();
        if !key.is_empty() {
            table.constraints.push(format!(
                "/* keying index */ unique({})",
                key.into_iter()
                    .map(|s| s.name.to_string())
                    .reduce(|a, b| format!("{},{}", a, b))
                    .unwrap()
            ));
        }

        tables.insert(table_name, table);
    }

    let mut tsort: TopologicalSort<&str> = TopologicalSort::new();
    for table in tables.values() {
        tsort.insert(table.table_name.as_str());

        for dep in table.dependencies.iter() {
            tsort.add_dependency(dep.as_str(), table.table_name.as_str());
        }
    }

    // this must be a stable hash function, so we very explicitly want to use a SipHasher with
    // known parameters
    #[allow(deprecated)]
    let mut signature_hasher = std::hash::SipHasher::new();
    use std::hash::{Hash, Hasher};

    let mut queries = vec![];

    loop {
        let mut table_list = tsort.pop_all();
        if table_list.is_empty() {
            break;
        }
        // bring into stable ordering
        table_list.sort();

        for table_name in table_list.into_iter() {
            let table = tables.get(table_name).unwrap();
            let create_sql = table.build_creation_query();

            table_name.hash(&mut signature_hasher);
            create_sql.hash(&mut signature_hasher);
            queries.push(create_sql);
        }
    }

    if !tsort.is_empty() {
        panic!("Cycle detected in dependency keys!")
    }

    // TODO: generate index schemas here

    DatabaseSchema {
        signature: signature_hasher.finish(),
        queries,
    }
}
