use crate::{
    db::{StatementContext, StatementRow},
    schema::{AssocData, AssocDomain, AssocMap, AssocRange, Entity, EntityVisitor, Relation},
    DBResult,
};

// trait implementations for Datum
mod datum_common;
// trait implementations for DatumList
mod datum_list;

// ----------------------------------------------------------------------
// Datum and related types
// ----------------------------------------------------------------------

/// Represents a data field in an Entity.
pub trait Datum: Clone + std::fmt::Debug {
    fn sql_type() -> &'static str;

    fn debug_field(&self, field: &'static str, fmt: &mut std::fmt::DebugStruct)
    where
        Self: Sized,
    {
        fmt.field(field, self);
    }

    fn bind_to(&self, _stmt: &mut StatementContext, index: i32);
    fn build_from(_adata: AssocData, _stmt: &mut StatementRow, _index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        unreachable!()
    }
    fn update_adata(&mut self, _adata: AssocData) {}

    fn accept_entity_visitor(_: &mut impl EntityVisitor) {}
    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_bare_field::<Self>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_bare_field::<Self>(self);
    }
}

/// Marker trait for 'concrete' datums, i.e. those with a static lifetime. This separation is used
/// to ensure [`QueryEquivalent`] can be implemented in the current Rust type system.
pub trait ConcreteDatum: 'static + Datum {}

/// Visitor for allowing for type-specific behaviour
pub trait DatumDiscriminator {
    fn visit_entity_id<E: Entity>(&mut self);
    fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(&mut self);
    fn visit_bare_field<T: Datum>(&mut self);
    fn visit_assoc_map<E: Entity>(&mut self);
    fn visit_assoc_domain<R: Relation>(&mut self);
    fn visit_assoc_range<R: Relation>(&mut self);
}

pub trait DatumDiscriminatorRef {
    fn visit_entity_id<E: Entity>(&mut self, _: &E::ID);
    fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(&mut self, _: &T);
    fn visit_bare_field<T: Datum>(&mut self, _: &T);
    fn visit_assoc_map<E: Entity>(&mut self, _: &AssocMap<E>);
    fn visit_assoc_domain<R: Relation>(&mut self, _: &AssocDomain<R>);
    fn visit_assoc_range<R: Relation>(&mut self, _: &AssocRange<R>);
}

/// A fixed-length list of EntityDatums, usually a tuple.
pub trait DatumList: Clone {
    fn accept(&self, visitor: &mut impl DatumVisitor);

    const LEN: usize;
}

/// A list of concrete datums.
pub trait ConcreteDatumList: DatumList + Clone {
    fn build_equivalent<'l>(
        from: impl Iterator<Item = &'l str>,
    ) -> Option<impl QueryEquivalentList<Self> + 'l>;
}

/// A walker for a DatumList instance.
pub trait DatumVisitor {
    fn visit<ED: Datum>(&mut self, datum: &ED);
}

/// A forward-edge mapping between datums where one can stand in for the other during queries.
///
/// This is purely a marker trait, since the equivalence is based on how sqlite handles things.
/// Because this is a forward-edge mapping, if type `X` implements `QueryEquivalent<Y>`, it means `X` can
/// stand in for `Y` during a query. For example, the following should be true:
/// 1. for datum `T`, `&T` should implement `QueryEquivalent<T>`
///     - rationale: you should be able to query for something matching a value without passing an
///     owned value
/// 2. for datum `T`, `StringQuery` should implement `QueryEquivalent<T>`
///     - rationale: `StringQuery` explicitly shoves off equality checking to sqlite, which means
///     you should be able to use it in any query place.
///
/// Note that this is abstracted across [`ConcreteDatum`] and not [`Datum`]; this is to avoid
/// issues with [`StringQuery`] -- which is not concrete -- having two implementations of
/// `QueryEquivalent<StringQuery>`: one from the reflexive implementation and one from (2) above.
pub trait QueryEquivalent<T: ConcreteDatum>: Datum {}

// Every type is query-equivalent to itself.
impl<T: ConcreteDatum> QueryEquivalent<T> for T {}
// Every reference type is query-equivalent to the non-referenced type.
impl<'l, T: ConcreteDatum> QueryEquivalent<T> for &'l T {}

/// Used if you wish to use sqlite's built-in string equivalence for a query.
#[derive(Clone, Debug)]
pub struct StringQuery<'l>(pub &'l str);

/// A list version of [`QueryEquivalent`].
pub trait QueryEquivalentList<T: DatumList>: DatumList {}
