use std::{fmt::Debug, hash::Hash};

use crate::{
    db::{Connection, StatementRow},
    schema::datum::Datum,
    DBResult,
};

use super::datum::{ConcreteDatum, ConcreteDatumList, QueryEquivalentList};

pub(crate) mod helpers;

/// Integral identifier for an entity.
pub trait EntityID: 'static + PartialEq + Hash + PartialOrd + Debug + Copy + ConcreteDatum {
    type Entity: Entity<ID = Self>;

    /// Construct from a raw integer.
    ///
    /// Do not use this unless you are _very_ certain it is what you want.
    fn from_raw(id: i64) -> Self;
    fn into_raw(self) -> i64;
}

// ----------------------------------------------------------------------
// EntityPart and related types
// ----------------------------------------------------------------------

/// A single data field in an Entity, automatically declared and derived as part of `#[derive(Entity)]`.
pub trait EntityPart: Default + Clone + 'static {
    type Datum: ConcreteDatum;
    type Entity: Entity;

    fn part_name() -> &'static str;
    fn placeholder() -> &'static str;
    fn unique() -> bool;
    fn desc() -> Option<&'static str>;

    fn get_datum(from: &Self::Entity) -> &Self::Datum;
}

/// Visitor for traversing all `EntityPart`s in an `Entity` or `EntityPartList`.
pub trait EntityPartVisitor {
    fn visit<EP: EntityPart>(&mut self) {}
    fn visit_datum<EP: EntityPart>(&mut self, _datum: &EP::Datum) {}
    fn visit_datum_mut<EP: EntityPart>(&mut self, _datum: &mut EP::Datum) {}
}

/// List of EntityParts.
pub trait EntityPartList: 'static {
    type Entity: Entity;
    type DatumList: ConcreteDatumList + QueryEquivalentList<Self::DatumList>;

    type ListHead: EntityPart;
    type ListTail: EntityPartList<Entity = Self::Entity>;
    const IS_EMPTY: bool = false;

    fn build_datum_list(conn: &Connection, stmt: &mut StatementRow) -> DBResult<Self::DatumList>;

    fn accept_part_visitor(_: &mut impl EntityPartVisitor);
    fn accept_part_visitor_ref(datum_list: &Self::DatumList, _: &mut impl EntityPartVisitor);
}

// trait implementations for EntityPartList
mod part_list;

pub use part_list::EmptyList;

// ----------------------------------------------------------------------
// Entity and related types
// ----------------------------------------------------------------------

/// A single database entity, aka an object type that gets its own table.
pub trait Entity: 'static + std::fmt::Debug {
    type Parts: EntityPartList;
    type Keys: EntityPartList;
    type ID: EntityID<Entity = Self> + EntityPart<Datum = Self::ID, Entity = Self>;

    fn build(values: <Self::Parts as EntityPartList>::DatumList) -> Self;

    fn entity_name() -> &'static str;
    fn accept_part_visitor(visitor: &mut impl EntityPartVisitor);
    fn accept_part_visitor_ref(&self, visitor: &mut impl EntityPartVisitor);
    fn accept_part_visitor_mut(&mut self, _: &mut impl EntityPartVisitor);
}

/// Visitor for traversing all `Entity`s in a container.
pub trait EntityVisitor {
    fn visit<E: Entity>(&mut self);
}
